#include "XC2dWaveprop.c"

// dimensional information ( t,x,y,z,s, note that x,y,z info are after tapering/padding)
int   XC2dAcouMod_nt,XC2dAcouMod_nx,XC2dAcouMod_nz,XC2dAcouMod_ns;
float XC2dAcouMod_ot,XC2dAcouMod_ox,XC2dAcouMod_oz;
float XC2dAcouMod_dt,XC2dAcouMod_dx,XC2dAcouMod_dz;
float *XC2dAcouMod_sdep, *XC2dAcouMod_rdep;
int *XC2dAcouMod_isdep, *XC2dAcouMod_irdep;
//tapering/padding information ( in x,y,z directions) 
int   XC2dAcouMod_nxtape,XC2dAcouMod_nztape;
int   XC2dAcouMod_nxt,XC2dAcouMod_nzt;
float *XC2dAcouMod_vel;
float *XC2dAcouMod_veldtinvdx2;
float XC2dAcouMod_vmax;
//incremental z range of wavefield to be calculated as t increases
int   XC2dAcouMod_izmax_inc;

#define zxloc(iz,ix) ((iz)+(ix)*XC2dAcouMod_nz)
#define Errmsg(str)  {fprintf(stderr,str);return 1;}
#define Errmsg2(str,str1)  {fprintf(stderr,str,str1);return 1;}
#define Errmsg3(str,str1,str2)  {fprintf(stderr,str,str1,str2);return 1;}

int XC2dAcouMod_Init( int ntin,  int nxin, int nzin, int nsin,  \
                      float otin,float oxin,float ozin,          \
                      float dtin,float dxin,float dzin,          \
                      int nxtapein, int nztapein,             \
                      float* sdepin, float *rdepin, float* velin,           \
                      int needwvmvin, int jtmvin );
int XC2dAcouMod_op(int tsign, int is, float *src, float *dat, float *wvmv);
//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
int XC2dAcouMod_Init( int ntin,  int nxin, int nzin, int nsin,  \
                      float otin,float oxin,float ozin,          \
                      float dtin,float dxin,float dzin,          \ 
                      int nxtapein, int nztapein,             \ 
                      float* sdepin, float *rdepin, float* velin,           \
                      int needwvmvin, int jtmvin )
{
    int ix,iy,iz,stat;
    float temp,v,vmax;
    // take input parameters
    XC2dAcouMod_nt=ntin;
    XC2dAcouMod_nx=nxin;
    XC2dAcouMod_nz=nzin;
    XC2dAcouMod_ns=nsin;

    XC2dAcouMod_ot=otin;
    XC2dAcouMod_ox=oxin;
    XC2dAcouMod_oz=ozin;

    XC2dAcouMod_dt=dtin;
    XC2dAcouMod_dx=dxin;
    XC2dAcouMod_dz=dzin;

    XC2dAcouMod_nxtape=nxtapein;
    XC2dAcouMod_nztape=nztapein;

    XC2dAcouMod_sdep=sdepin;
    XC2dAcouMod_rdep=rdepin;
    XC2dAcouMod_vel=velin;

    // determine if padded region is reasonable
    if (nxin<=2*nxtapein)
       WpErrmsg("Error, nx less or equal than twice of padded x reigon\n")
    if (nzin<=2*nztapein)
       WpErrmsg("Error, nz less or equal than twice of padded z reigon\n")

    // determine if dx==dz
    if (  XC2dAcouMod_dz!=XC2dAcouMod_dx )
       Errmsg("ERROR, dz not equal to dx, exit!!!\n")
    else 
    {
       // produce (vel*dt/dx)**2 and get vmax/izmax_inc on the way
       vmax=0.;
       XC2dAcouMod_isdep=(int*)malloc(nxin*sizeof(int));
       XC2dAcouMod_irdep=(int*)malloc(nxin*sizeof(int));
       XC2dAcouMod_veldtinvdx2=(float*)malloc(nxin*nzin*sizeof(float));
       for (ix=0;ix<nxin;ix++)
          for (iz=0;iz<nzin;iz++)
          {
             v=XC2dAcouMod_vel[ zxloc(iz,ix) ];
             temp= v * XC2dAcouMod_dt/XC2dAcouMod_dx;
             XC2dAcouMod_veldtinvdx2[ zxloc(iz,ix) ]=temp*temp;
             if ( v  > vmax)
                vmax = v;
             if (temp>0.86)
                Errmsg2("ERROR, stability condition for wavepropagation is violated, temp=%f which is larger than 0.86\n",temp)
          }
       XC2dAcouMod_izmax_inc= vmax*XC2dAcouMod_dt/XC2dAcouMod_dz + 1;
    }
    // initilize wave propagation module
    int freesurf=0;
    long memcnt=0; 
    stat=XC2dWaveprop_Init(ntin,nxin,nzin,dtin, dxin, dzin, nxtapein,nztapein,needwvmvin,jtmvin, freesurf, &memcnt);
    if (stat==1)
       Errmsg("Error in XC2dWaveprop_Init, program terminated!\n")

    return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
int XC2dAcouMod_op(int tsign, int is, float *src, float *dat, float *wvmv)
{
  float *sdep,*rdep;
  int ix;
  int nx;
  int izmax_inc,stat;
  float dz;

  //VerbMsg("Calculating sdep\n")
  sdep=&XC2dAcouMod_sdep[is*XC2dAcouMod_nx];
  rdep=&XC2dAcouMod_rdep[is*XC2dAcouMod_nx];
 
  nx=XC2dAcouMod_nx;
  dz=XC2dAcouMod_dz;
  // convert float depth (sdep, rdep) into integer index depth(isdep, irdep) 
 
  //VerbMsg("Calculating isdep\n")
  for (ix=0;ix<nx;ix++)
     XC2dAcouMod_isdep[ ix ]=(sdep[ ix ]-XC2dAcouMod_oz)/dz+XC2dAcouMod_nztape;

  //VerbMsg("Calculating irdep\n")
  for (ix=0;ix<nx;ix++)
     XC2dAcouMod_irdep[ ix ]=(rdep[ ix ]-XC2dAcouMod_oz)/dz+XC2dAcouMod_nztape;


  // wave propgation
  izmax_inc=XC2dAcouMod_izmax_inc;
  stat=XC2dWaveprop_op( tsign, XC2dAcouMod_isdep, XC2dAcouMod_irdep, src, XC2dAcouMod_veldtinvdx2, izmax_inc, dat, wvmv );

  if (stat==1) 
     Errmsg("Error in XC2dWaveprop_op, program terminated!\n")   
  
  return 0;
}
