#include "stdlib.h"
#include "math.h"
#include "string.h"
#include <sep_pars_external.h>
#include <seplib.h>
#include <sep3dc.h>
#include "fftw3.h"
#include "XC2dAcouMod.c"

int main(int argc, char** argv)
{
    int ndim,i,n;
    float o,d;
    char temp_char[256];
    int   nt, nx, nz, ns;
    float ot, ox, oz;
    float dt, dx, dz;
    int   it, ix, iz, is;
    int   nxtape,nztape;
    int   needwvmv,jtmv;
    int   stat,tsign;
    float *vel, *src, *data, *sdep, *rdep, *wvmv;
    sep3d outfile; 
    time_t time1,time2;

    initpar(argc,argv);
    // read t,x,y from source file
   
    if(0!=sep_get_number_data_axes("in",&ndim))
       seperr("unable to get dimension from src!\n");
    
    for(i=1;i<=ndim;i++)
    {
       if(0!=sep_get_data_axis_par("in",&i,&n,&o,&d,temp_char))
       seperr("unable to get parameters from src!\n");
       if(i==1)
       {
          nt=n;
          dt=d;
          ot=o;
       }
       else if(i==2)
       {
          nx=n;
          dx=d;
          ox=o;
       }
       if (ndim==2) 
          ns=1; 
       else if(i==3)
          ns=n;
    }

    // read z info from vel
    if(0!=sep_get_number_data_axes("vel",&ndim))
       seperr("unable to get dimension from vel file!\n");
    for(i=1;i<=ndim;i++)
    {
       if(0!=sep_get_data_axis_par("vel",&i,&n,&o,&d,temp_char))
       seperr("unable to get parameters from vel file!\n");
       if(i==1)
       {
          nz=n;
          dz=d;
          oz=o;
       }
       else if(i==2)
       {
          if (nx!=n)
          Errmsg3("nx in vel %d different from nx %d in src, program terminated!\n",nx,n)
          if (dx!=d)
          Errmsg3("dx in vel %f different from dx %f in src, program terminated!\n",dx,d)
          if (ox!=o)
          Errmsg3("ox in vel %f different from ox %f in src, program terminated!\n",ox,o)
       }
    }

    // readin other parameters
    if (0==getch("nxtape","i", &nxtape)) nxtape=50;
    if (0==getch("nztape","i", &nztape)) nztape=50;
    if (0==getch("needwvmv","i", &needwvmv)) needwvmv=0;
    if (0==getch("jtmv","i", &jtmv)) jtmv=10;
    if (0==getch("tsign","i", &tsign)) tsign=1;

    if (nx<=2*nxtape)
       Errmsg3("Error, nx %d less or equal than twice of padded x reigon %d \n",nx,nxtape);
    if (nz<=2*nztape)
       Errmsg3("Error, nz %d less or equal than twice of padded z reigon %d \n",nz,nztape);

    // allocate memories

    vel=(float*)malloc(nx*nz*sizeof(float));
    src=(float*)malloc(nx*nt*ns*sizeof(float));
    data=(float*)malloc(nx*nt*ns*sizeof(float));
    sdep=(float*)malloc(nx*ns*sizeof(float));
    rdep=(float*)malloc(nx*ns*sizeof(float));

    for (i=0;i<nx*nt*ns;i++)
       data[i]=0; 
    // read in various data files into arrays
    if ((nx*nt*ns*4)!=sreed("in",src,nx*nt*ns*4))
       Errmsg2("Error read in src file of size %d \n",nx*nt*ns*4);

    if ((nx*nz*4)!=sreed("vel",vel,nx*nz*4))
       Errmsg2("Error read in vel file of size %d \n", nx*nz*4);

    FILE *fp;
    fp=fopen("sdep","r");
    if (fp!=NULL)
    {
       fclose(fp);
       if ((nx*ns*4)!=sreed("sdep",sdep,nx*ns*4))
          seperr("Error read in sdep file\n");
    }
    else
    {
       for (i=0;i<nx*ns;i++)
          sdep[i]=2*dz;
    }

    fp=fopen("rdep","r");
    if (fp)
    {
       fclose(fp);
       if ((nx*ns*4)!=sreed("rdep",rdep,nx*ns*4))
          seperr("Error read in rdep file\n");
    }
    else
    {
       for (i=0;i<nx*ns;i++)
          rdep[i]=2*dz;
    }

    if (needwvmv>0) 
       wvmv=(float*)malloc(nx*nz*((nt-1)/jtmv+1)*sizeof(float));
     
    // Initialize wave propagation 
    VerbMsg("start initialize XC2dAcouMod_Init\n")
    stat=XC2dAcouMod_Init( nt,nx,nz,ns,       \
                           ot,ox,oz,          \
                           dt,dx,dz,          \
                           nxtape,nztape, \
                           sdep, rdep, vel,      \
                           needwvmv,jtmv );

    if (stat==1)
       seperr("Error in XC2dAcouMod_Init,program terminated!\n");
    
    // run wave propagation 
    VerbMsg("start modeling in XC2dAcouMod_op \n")
    time1=time(NULL);
    for (is=0;is<ns;is++) 
       stat=XC2dAcouMod_op(tsign,is,&src[is*nx*nt],&data[is*nx*nt],wvmv);
    time2=time(NULL);

    fprintf(stderr,"used time for 1 shot is %ld seconds\n", time2-time1);
    if (stat==1)
       seperr("Error in XC2dAcouMod_op,program terminated!\n");

    //output recorded data
    VerbMsg("start output recorded data \n")
    ndim=3;
    if(0!=init_sep3d_par(&outfile,"OUTPUT","FLOAT","REGULAR",ndim,0))
       seperr("unable to write outfile!\n");
    outfile.o[0]=ot;
    outfile.d[0]=dt;
    outfile.n[0]=nt;
    outfile.o[1]=ox;
    outfile.d[1]=dx;
    outfile.n[1]=nx;
    outfile.o[2]=0;
    outfile.d[2]=1;
    outfile.n[2]=ns;
    sep3d_set_sep3d(&outfile);
    if(0!=sep3dc_write_description("data",&outfile))
       seperr("unable to output parameters for data\n");

    if( (nt*nx*ns*4)!=srite("data",data,4*nt*nx*ns) )
       seperr("trouble writing out data\n");

    //output wave movie if necessary
    if (needwvmv>0)
    {
       ndim=3;
       VerbMsg("start output wave movie \n")
       if(0!=init_sep3d_par(&outfile,"OUTPUT","FLOAT","REGULAR",ndim,0))
          seperr("unable to write outfile!\n");
       outfile.o[0]=oz;
       outfile.d[0]=dz;
       outfile.n[0]=nz;
       outfile.o[1]=ox;
       outfile.d[1]=dx;
       outfile.n[1]=nx;
       outfile.o[2]=ot;
       outfile.d[2]=dt*jtmv;
       outfile.n[2]=nt/jtmv;
       sep3d_set_sep3d(&outfile);
       VerbMsg("start output wave movie header \n" )
       if(0!=sep3dc_write_description("wvmv",&outfile))
          seperr("unable to output parameters for wvmv\n");

       VerbMsg("start output wave movie data body \n" )
       if( (nx*nz*(nt/jtmv)*4)!=srite("wvmv",wvmv,4*nx*nz*(nt/jtmv)) )
          seperr("trouble writing out wvmv\n");
       VerbMsg("done output wave movie data body \n" )
    }
    return 0;
}
