#include "stdlib.h"
#include "math.h"
#include "string.h"
#include <sep_pars_external.h>
#include <seplib.h>
#include <sep3dc.h>
#include "fftw3.h"

int main(int argc, char** argv)
{
    int ndim,i,n;
    float o,d;
    char temp_char[256];
    int   nt, nx, ns;
    float ot, ox, os;
    float dt, dx, ds;
    int   it, i1, i2, itlag;
    int   nxtape,nztape;
    int   ntlag;
    int   maxidx;
    float *dat1, *dat2, *xcorr, *maxt;
    float *tr1, *tr2;
    float maxval,thres;
    sep3d outfile; 
    time_t time1,time2;

    initpar(argc,argv);
    // read t,x,y from source file
   
    if(0!=sep_get_number_data_axes("in",&ndim))
       seperr("unable to get dimension from src!\n");
    
    for(i=1;i<=ndim;i++)
    {
       if(0!=sep_get_data_axis_par("in",&i,&n,&o,&d,temp_char))
       seperr("unable to get parameters from src!\n");
       if(i==1)
       {
          nt=n;
          dt=d;
          ot=o;
       }
       else if(i==2)
       {
          nx=n;
          dx=d;
          ox=o;
       }
       if (ndim==2) 
          ns=1; 
       else if(i==3)
       {
          ns=n;
          ds=d;
          os=o;
       }
    }

    fprintf(stderr,"start reading dat2 header!\n");
    // read z info from vel
    if(0!=sep_get_number_data_axes("dat2",&ndim))
       seperr("unable to get dimension from vel file!\n");
    for(i=1;i<=ndim;i++)
    {
       if(0!=sep_get_data_axis_par("dat2",&i,&n,&o,&d,temp_char))
       seperr("unable to get parameters from vel file!\n");
       if(i==1)
       {
          if (nt!=n)
          fprintf(stderr,"nt in dat2 %d different from nt %d in dat1, program terminated!\n",n,nt);
          if (dt!=d)
          fprintf(stderr,"dt in dat2 %f different from dt %f in dat1, program terminated!\n",d,dt);
          if (ot!=o)
          fprintf(stderr,"ot in dat2 %f different from ot %f in dat1, program terminated!\n",o,ot);
       }
       else if(i==2)
       {
          if (nx!=n)
          fprintf(stderr,"nx in dat2 %d different from nx %d in dat1, program terminated!\n",n,nx);
          if (dx!=d)
          fprintf(stderr,"dx in dat2 %f different from dx %f in dat1, program terminated!\n",d,dx);
          if (ox!=o)
          fprintf(stderr,"ox in dat2 %f different from ox %f in dat1, program terminated!\n",o,ox);
       }
       else if(i==3 && ns!=1)
       {
          if (ns!=n)
          fprintf(stderr,"ns in dat2 %d different from ns %d in dat1, program terminated!\n",n,ns);
          if (ds!=d)
          fprintf(stderr,"ds in dat2 %f different from ds %f in dat1, program terminated!\n",d,ds);
          if (os!=o)
          fprintf(stderr,"os in dat2 %f different from os %f in dat1, program terminated!\n",o,os);
       }
    }
    // readin other parameters
    if (0==getch("ntlag","i", &ntlag)) ntlag=50;
    if (0==getch("thres","f", &thres)) thres=0.0005;

    dat1=(float*)malloc(nx*nt*ns*sizeof(float));
    dat2=(float*)malloc(nx*nt*ns*sizeof(float));
    tr1=(float*)malloc(nt*sizeof(float));
    tr2=(float*)malloc(nt*sizeof(float));
    xcorr=(float*)malloc(nx*ns*(2*ntlag+1)*sizeof(float));
    maxt=(float*)malloc(nx*ns*sizeof(float));

    // read in various data files into arrays
    if ((nx*nt*ns*4)!=sreed("in",dat1,nx*nt*ns*4))
       fprintf(stderr,"Error read in dat1 file of size %d \n",nx*nt*ns*4);

    if ((nx*nt*ns*4)!=sreed("dat2",dat2,nx*nt*ns*4))
       fprintf(stderr,"Error read in dat2 file of size %d \n",nx*nt*ns*4);

    fprintf(stderr,"start calculaation\n");
    // run wave propagation 
   for (i2=0;i2<nx*ns;i2++)
   {
      for (i1=0;i1<2*ntlag+1;i1++)
         xcorr[i1]=0;
      for (i1=0;i1<nt;i1++)
         tr1[i1]=dat1[i1+i2*nt];
      for (itlag=-ntlag;itlag<1;itlag++)
      {
         for (i1=0;i1<nt;i1++)
            tr2[i1]=0;
         for (i1=0;i1<nt+itlag;i1++)
            tr2[i1]=dat2[i1-itlag+i2*nt];
         for (i1=0;i1<nt;i1++)
            xcorr[itlag+ntlag]+=tr1[i1]*tr2[i1];
      }
      for (itlag=1;itlag<ntlag;itlag++)
      {
         for (i1=0;i1<nt;i1++)
            tr2[i1]=0;
         for (i1=itlag;i1<nt;i1++)
            tr2[i1]=dat2[i1-itlag+i2*nt];
         for (i1=0;i1<nt;i1++)
            xcorr[itlag+ntlag]+=tr1[i1]*tr2[i1];
      }
      maxval=thres;
      maxidx=0;
      maxt[i2]=0;
      for (itlag=0;itlag<2*ntlag+1;itlag++)
      {
         if (xcorr[itlag]>maxval)
         {
            maxval=xcorr[itlag];
            maxt[i2]=(itlag-ntlag)*dt;
         }
      }
   }


    ndim=2;
    if(0!=init_sep3d_par(&outfile,"OUTPUT","FLOAT","REGULAR",ndim,0))
       seperr("unable to write outfile!\n");
    outfile.o[0]=ox;
    outfile.d[0]=dx;
    outfile.n[0]=nx;
    outfile.o[1]=os;
    outfile.d[1]=ds;
    outfile.n[1]=ns;
    sep3d_set_sep3d(&outfile);
    if(0!=sep3dc_write_description("maxt",&outfile))
       seperr("unable to output parameters for imag\n");

    if( (ns*nx*4)!=srite("maxt",maxt,4*ns*nx) )
       seperr("trouble writing out maxt\n");

    return 0;
}
