int XC2dWI_m2v(float *m, float *v, int nm, int nv)
{
   int i;
   if (nm==nv)
      for (i=0;i<nm;i++)
         v[i]=m[i];
   else 
   {
      fprintf(stderr,"Error in XC3dWI_m2v, nm!=nv\n");
      return 1;
   } 

   return 0; 
}

int XC2dWI_caldirecm(double *grad1in, double *grad2in, double *direcmin, double *direcmout, int sz )
{
  double tmp1, tmp2;
  double res1, res2, res3;
  int i;

  res1=0;
  res2=0;
  res3=0;

  for (i=0;i<sz;i++)
  {
     tmp1=grad1in[i];
     tmp2=grad2in[i];

     res1+=tmp1*tmp1;
     res2+=tmp1*tmp2;
     res3+=tmp2*tmp2;
  }
  
  double beta;
  if (res1==0.) 
     beta=0.;
  else
     beta = (res3-res2) / res1;
   
  if (beta<0.)
     beta = 0.;

  for (i=0;i<sz;i++)
     direcmout[i]=grad2in[i]+beta*direcmin[i];
     //direcmout[i]=grad2in[i];
     //direcmout[i]=grad2in[i]+beta*direcmin[i];

  return 0;
}


/* nonlinear conjugate gradient with preconditioning:

beta= ( (g_n/p_n)(g_n-g_(n-1)) )/ ( (g_(n-1)/p_(n-1))*g_(n-1) )
*/
int XC2dWI_caldirecmwithprecond(double *grad1in, double *grad2in, double *precond2in, double *denomin, double *denomout, double *direcmin, double *direcmout, int sz )
{
  double tmp1, tmp2;
  double res1, res2, res3;
  int i;

  res1=0;
  res2=0;
  res3=0;

  double maxval=0;
  for (i=0;i<sz;i++)
     if (abs(grad2in[i])>maxval)
        maxval=abs(grad2in[i]);

  for (i=0;i<sz;i++)
  {
     tmp1=grad1in[i];
     tmp2=grad2in[i]/(precond2in[i]+maxval*0.01);

     res2+=tmp1*tmp2;
     res3+=tmp2*grad2in[i];
  }
  
  double beta;
  if (*denomin==0.) 
     beta=0.;
  else
     beta = (res3-res2) / *denomin;
   
  if (beta<0.)
     beta = 0.;

  for (i=0;i<sz;i++)
     direcmout[i]=grad2in[i]/(precond2in[i]+maxval*0.01)+beta*direcmin[i];

  *denomout=res3;

  return 0;
}

int XC2dWI_padgrad(double *gradin, double *gradout, int nx, int nz, int nxtape, int nztape )
{
   int ix,iz;
   
   for (ix=nxtape;ix<nx-nxtape;ix++)
   {
      for (iz=0;iz<nztape;iz++)
        gradout[iz+ix*nz]=gradin[nztape+ix*nz];

      for (iz=nz-nztape;iz<nz;iz++)
        gradout[iz+ix*nz]=gradin[nz-nztape-1+ix*nz];
   }

   for (ix=0;ix<nxtape;ix++)
      for (iz=0;iz<nz;iz++)
         gradout[iz+ix*nz]=gradin[iz+nxtape*nz];

   for (ix=nx-nxtape;ix<nx;ix++)
      for (iz=0;iz<nz;iz++)
         gradout[iz+ix*nz]=gradin[iz+(nx-nxtape-1)*nz];

  return 0;
}

int XC2dWI_padfgrad(float *gradin, float *gradout, int nx, int nz, int nxtape, int nztape )
{
   int ix,iz;
   
   for (ix=nxtape;ix<nx-nxtape;ix++)
   {
      for (iz=0;iz<nztape;iz++)
        gradout[iz+ix*nz]=gradin[nztape+ix*nz];

      for (iz=nz-nztape;iz<nz;iz++)
        gradout[iz+ix*nz]=gradin[nz-nztape-1+ix*nz];
   }

   for (ix=0;ix<nxtape;ix++)
      for (iz=0;iz<nz;iz++)
        gradout[iz+ix*nz]=gradin[iz+nxtape*nz];

   for (ix=nx-nxtape;ix<nx;ix++)
      for (iz=0;iz<nz;iz++)
        gradout[iz+ix*nz]=gradin[iz+(nx-nxtape-1)*nz];

  return 0;
}

// input mod has 3 components: vp^2, vp^2*(1+2*eps), vp^2*sqrt(1+2*delta)
// paramtype=1 means use vp^2 and vp^2(1+2*eps) as two model components
// paramtype=2 means use a^(-1)ln(vp^(-2)) and 1+2*eps as two model components
int XC2dWI_updatemod(float *mod, double *grad, float alpha, float *newmod, int n, float vmin, float vmax, int paramtype, float dt, float dx )
{
   int i;
   double val1,val2;
   float a;
   a=1.5;
   double tmpval;
   if (vmax<=vmin)
      WpErrmsg("vmax <= vmin in XC2dWI_updatemod\n")
   if(paramtype==1) 
   {
      for (i=0;i<n;i++)
      {   
         newmod[i+2*n]=mod[i+2*n]/mod[i];
         newmod[i+n]=mod[i+n]+alpha*grad[i+n];
         newmod[i]=mod[i]+alpha*grad[i];
         newmod[i+2*n]*=newmod[i];
         if (newmod[i]>vmax)
            newmod[i]=vmax;
         
         if (newmod[i]<vmin)
            newmod[i]=vmin;
       
         //next make sure eps>=delta  
         val1=newmod[i+n]/newmod[i];
         val2=newmod[i+2*n]/newmod[i]; 
         if (val1<val2*val2)
            newmod[i+n]=(val2*val2+1.e-5)*newmod[i];
       }
   } 
   else if(paramtype==2)
   {
      //make sure eps>=delta by zero out eps gradient 
      //for (i=0;i<n;i++)
      //   if (grad[i+n]<0)
      //      grad[i+n]=0;
      for (i=0;i<n;i++)
      {   
         //transform back to (vp*dt/dx)^2, 1+2*eps, sqrt(1+2*delta)
         newmod[i+2*n]=mod[i+2*n]/mod[i];
         newmod[i+n]=mod[i+n]/mod[i];
         newmod[i]=mod[i];
         // update mod using parameterization 
         tmpval=1./a*log((dt*dt/(newmod[i]*dx*dx)));
         tmpval+=grad[i]*((double)alpha);
         newmod[i]=dt*dt/(exp((a*tmpval))*dx*dx);
         newmod[i+n]+=alpha*grad[i+n];
         newmod[i+n]*=newmod[i];
         newmod[i+2*n]*=newmod[i];
         if (newmod[i]>vmax)
            newmod[i]=vmax;
         
         if (newmod[i]<vmin)
            newmod[i]=vmin;
       
         //next make sure eps>=delta  
         
         val1=newmod[i+n]/newmod[i];
         val2=newmod[i+2*n]/newmod[i]; 
         if (val1<val2*val2)
            newmod[i+n]=(val2*val2+1.e-5)*newmod[i];
         
       }
   } 
   return 0; 
}

int XC2dWI_equalepsdel(float *mod, float *newmod, int nz, int nx, int isdepmin, int isdepmax, int nztape, int nxtape)
{
   int ix,iz,idx;
   double val1,val2;
   double newval;
   for (ix=0;ix<nx;ix++)
      for (iz=0;iz<nz;iz++)
      {   
         idx=iz+ix*nz;
         newmod[idx]=mod[idx];
         newmod[idx+nz*nx]=mod[idx+nz*nx];
         newmod[idx+2*nz*nx]=mod[idx+2*nz*nx];
      }
   for (ix=0;ix<nx;ix++)
      for (iz=max(isdepmin,0);iz<min(isdepmax+1,nz);iz++)
      {   
         idx=iz+ix*nz;
         val1=mod[idx+nz*nx]/mod[idx];
         val2=mod[idx+2*nz*nx]/mod[idx]; 
         //newval=0.5*(val1+val2*val2);   
         newval=val2*val2;   
         newmod[idx+nz*nx]=(newval+1.e-5)*newmod[idx];
         //newmod[idx+2*nz*nx]=sqrt(newval)*newmod[idx];
         if (iz==isdepmin)
         {
            idx=iz-1+ix*nz;
            val1=mod[idx+nz*nx]/mod[idx];
            val2=mod[idx+2*nz*nx]/mod[idx]; 
            //newval=0.5*(val1+val2*val2);   
            newval=val2*val2;   
            newmod[idx+nz*nx]=(newval+1.e-5)*newmod[idx];
            //newmod[idx+2*nz*nx]=sqrt(newval)*newmod[idx];
            idx=iz-2+ix*nz;
            val1=mod[idx+nz*nx]/mod[idx];
            val2=mod[idx+2*nz*nx]/mod[idx]; 
            //newval=0.5*(val1+val2*val2);   
            newval=val2*val2;   
            newmod[idx+nz*nx]=(newval+1.e-5)*newmod[idx];
            //newmod[idx+2*nz*nx]=sqrt(newval)*newmod[idx];
            idx=iz-3+ix*nz;
            val1=mod[idx+nz*nx]/mod[idx];
            val2=mod[idx+2*nz*nx]/mod[idx]; 
            newval=0.5*(val1+val2*val2);   
            //newval=0.75*val1+0.25*val2*val2;   
            newmod[idx+nz*nx]=(newval+1.e-5)*newmod[idx];
            //newval=0.25*val1+0.75*val2*val2; 
            newmod[idx+2*nz*nx]=sqrt(newval)*newmod[idx];
         }
         if (iz==isdepmax)
         {
            idx=iz+1+ix*nz;
            val1=mod[idx+nz*nx]/mod[idx];
            val2=mod[idx+2*nz*nx]/mod[idx]; 
            //newval=0.5*(val1+val2*val2);   
            newval=val2*val2;   
            newmod[idx+nz*nx]=(newval+1.e-5)*newmod[idx];
            //newmod[idx+2*nz*nx]=sqrt(newval)*newmod[idx];
            idx=iz+2+ix*nz;
            val1=mod[idx+nz*nx]/mod[idx];
            val2=mod[idx+2*nz*nx]/mod[idx]; 
            //newval=0.5*(val1+val2*val2);   
            newval=val2*val2;   
            newmod[idx+nz*nx]=(newval+1.e-5)*newmod[idx];
            //newmod[idx+2*nz*nx]=sqrt(newval)*newmod[idx];
            idx=iz+3+ix*nz;
            val1=mod[idx+nz*nx]/mod[idx];
            val2=mod[idx+2*nz*nx]/mod[idx]; 
            //newval=0.75*val1+0.25*val2*val2;   
            newval=0.5*(val1+val2*val2);   
            newmod[idx+nz*nx]=(newval+1.e-5)*newmod[idx];
            //newval=0.25*val1+0.75*val2*val2;   
            newmod[idx+2*nz*nx]=sqrt(newval)*newmod[idx];
         }
       }
   return 0; 
}

int XC2dWI_rndbndv1(float *vel, int nx, int nz, int nxtape, int nztape, float sigma, float alpha_mean, float alpha_sigma, float  alpha_maxV, int j_rndbnd, float minV, float maxV)
{
   int i,ix,iz;
   int i1,ix1,iz1;
   if (maxV<=minV)
      WpErrmsg("vmax <= vmin in XC3dWI_updatev\n")
    // first pad all boundaries by extending values
    for (ix=0;ix<nxtape;ix++)
       for (iz=nztape;iz<nz-nztape;iz++)
       {
          vel[iz+ix*nz]=vel[iz+nxtape*nz];
          vel[iz+(nx-1-ix)*nz]=vel[iz+(nx-1-nxtape)*nz];
       }

    for (ix=0;ix<nx;ix++)
       for (iz=0;iz<nztape;iz++)
       {
          vel[iz+ix*nz]=vel[nztape+ix*nz];
          vel[(nz-1-iz)+ix*nz]=vel[(nz-1-nztape)+ix*nz];
       }

   // first adjust mean of boundary regions
   for (ix=0;ix<nxtape;ix++)
       for (iz=0;iz<nz;iz++)
       {
          vel[iz+ix*nz]*=exp(-alpha_mean*(nxtape-ix));
          vel[iz+(nx-1-ix)*nz]*=exp(-alpha_mean*(nxtape-ix));
       }

   for (ix=0;ix<nx;ix++)
      for (iz=0;iz<nztape;iz++)
      {
         vel[iz+ix*nz]*=exp(-alpha_mean*(nztape-iz));
         vel[(nz-1-iz)+ix*nz]*=exp(-alpha_mean*(nztape-iz));
      }

   // then adjust the var of boundary regions
   for (ix=0;ix<nxtape;ix++)
      for (iz=0;iz<nz;iz++)
      {
         i=iz+ix*nz;
         vel[iz+ix*nz]+=sigma*exp(alpha_sigma*(nxtape-ix))*((2.0*rand()/RAND_MAX)-1.0);
         if (vel[i]>maxV*exp(alpha_maxV*(nxtape-1-ix)))
             vel[i]=maxV*exp(alpha_maxV*(nxtape-1-ix));
         i=iz+(nx-1-ix)*nz;
         vel[i]+=sigma*exp(alpha_sigma*(nxtape-ix))*((2.0*rand()/RAND_MAX)-1.0);
         if (vel[i]>maxV*exp(alpha_maxV*(nxtape-1-ix)))
            vel[i]=maxV*exp(alpha_maxV*(nxtape-1-ix));
      }

   for (ix=0;ix<nx;ix++)
      for (iz=0;iz<nztape;iz++)
      {
         i=iz+ix*nz;
         vel[i]+=sigma*exp(alpha_sigma*(nztape-iz))*((2.0*rand()/RAND_MAX)-1.0);
         if (vel[i]>maxV*exp(alpha_maxV*(nztape-1-iz)))
            vel[i]=maxV*exp(alpha_maxV*(nztape-1-iz));
         i=(nz-1-iz)+ix*nz;
         vel[i]+=sigma*exp(alpha_sigma*(nztape-iz))*((2.0*rand()/RAND_MAX)-1.0);
         if (vel[i]>maxV*exp(alpha_maxV*(nztape-1-iz)))
            vel[i]=maxV*exp(alpha_maxV*(nztape-1-iz));
      }

   for (i=0;i<nx*nz;i++)
   {
      if (vel[i]<minV)
         vel[i]=minV;
   }
   
   for (ix=nxtape-1;ix>-1;ix-=j_rndbnd)
      for (iz=0;iz<nz;iz+=j_rndbnd)
      {
         for (ix1=0;ix1<j_rndbnd;ix1++)
            for (iz1=0;iz1<j_rndbnd;iz1++)
            {
                i=iz+ix*nz;
                i1=iz+iz1+(ix-ix1)*nz;
                if (((ix-ix1)>-1)&&((iz+iz1)<nz))
                   vel[i1]=vel[i];
                i=iz+(nx-1-ix)*nz;
                i1=iz+iz1+(nx-1-(ix-ix1))*nz;
                if (((ix-ix1)>-1)&&((iz+iz1)<nz))
                   vel[i1]=vel[i];
            }
      }

   for (ix=0;ix<nx;ix+=j_rndbnd)
      for (iz=nztape-1;iz>-1;iz-=j_rndbnd)
      {
         for (ix1=0;ix1<j_rndbnd;ix1++)
            for (iz1=0;iz1<j_rndbnd;iz1++)
            {
                i=iz+ix*nz;
                i1=iz-iz1+(ix+ix1)*nz;
                if (((iz-iz1)>-1)&&((ix+ix1)<nx))
                   vel[i1]=vel[i];
                i=(nz-1-iz)+ix*nz;
                i1=(nz-1-(iz-iz1))+(ix+ix1)*nz;
                if (((iz-iz1)>-1)&&((ix+ix1)<nx))
                   vel[i1]=vel[i];
            }
      }


   return 0; 
}

int XC2dWI_rndbndv2(float *vel, int nx, int nz, int nxtape, int nztape, float sigma, float alpha_mean, float alpha_sigma, float  alpha_maxV, int j_rndbnd, float minV, float maxV)
{
   int i,ix,iy,iz;
   int i1,ix1,iy1,iz1;
   int *msk;
   float maxsigma,wht;
   msk=(int*)malloc(nz*nx*sizeof(int));
   maxsigma=1000;

  //fprintf(stderr,"in rndbndv2, alpha_mean=%e, alpha_sigma=%e, alpha_maxV=%e, sigma=%e\n",alpha_mean,alpha_sigma, alpha_maxV, sigma);  
    // put 1 where velocity has been perturbed
    for (i=0;i<nx*nz;i++)
       msk[i]=0;
    
    for (ix=nxtape;ix<nx-nxtape;ix++)
       for (iz=nztape;iz<nz-nztape;iz++)
          msk[iz+ix*nz]=1;

    // then adjust the var of boundary region seeds
    for (ix=nxtape-1;ix>=0;ix-=j_rndbnd)
       for (iz=0;iz<nz;iz+=j_rndbnd)
       {
          i=iz+ix*nz;
          vel[i]+=(sigma*exp(alpha_sigma*(nxtape-ix)))*((2.0*rand()/RAND_MAX)-1.0);
          if (vel[i]>maxV/exp(alpha_maxV*(nxtape-ix)))
             vel[i]=maxV/exp(alpha_maxV*(nxtape-ix));
          msk[i]=1;
          i=iz+(nx-1-ix)*nz;
          vel[i]+=(sigma*exp(alpha_sigma*(nxtape-ix)))*((2.0*rand()/RAND_MAX)-1.0);
          if (vel[i]>maxV/exp(alpha_maxV*(nxtape-ix)))
             vel[i]=maxV/exp(alpha_maxV*(nxtape-ix));
          msk[i]=1;
       }

    // then adjust the var of the whole boundary regions using nearest neighbor (statistically in all three directions)
    for (ix=0;ix<nxtape;ix++)
       for (iz=0;iz<nz;iz++)
       {
          i=iz+(nxtape-1-ix)*nz;
          if (msk[i]==0)
          {
             iz1=((int)(iz/j_rndbnd))*j_rndbnd;
             ix1=((int)(ix/j_rndbnd))*j_rndbnd;
             if ((iz1+j_rndbnd)<nz)
             {
                wht=1.*(iz-iz1)/j_rndbnd;
                if (((1.0*rand())/RAND_MAX)<=wht)
                   iz1=iz1+j_rndbnd;
             }
             if ((ix1+j_rndbnd)<nxtape)
             {
                wht=1.*(ix-ix1)/j_rndbnd;
                if (((1.0*rand())/RAND_MAX)<=wht)
                   ix1=ix1+j_rndbnd;
             }
             i1=iz1+(nxtape-1-ix1)*nz;
             vel[i]=vel[i1];
          }

          i=iz+(nx-nxtape+ix)*nz;
          if (msk[i]==0)
          {
             iz1=((int)(iz/j_rndbnd))*j_rndbnd;
             ix1=((int)(ix/j_rndbnd))*j_rndbnd;
             if ((iz1+j_rndbnd)<nz)
             {
                wht=1.*(iz-iz1)/j_rndbnd;
                if (((1.0*rand())/RAND_MAX)<=wht)
                   iz1=iz1+j_rndbnd;
             }
             if ((ix1+j_rndbnd)<nxtape)
             {
                wht=1.*(ix-ix1)/j_rndbnd;
                if (((1.0*rand())/RAND_MAX)<=wht)
                   ix1=ix1+j_rndbnd;
             }
             i1=iz1+(nx-nxtape+ix1)*nz;
             vel[i]=vel[i1];
          }
       }

    // put 1 where velocity has been perturbed
    for (i=0;i<nx*nz;i++)
       msk[i]=0;

    for (ix=nxtape;ix<nx-nxtape;ix++)
       for (iz=nztape;iz<nz-nztape;iz++)
          msk[iz+ix*nz]=1;

    // then adjust the var of boundary region seeds
    for (ix=0;ix<nx;ix+=j_rndbnd)
       for (iz=nztape-1;iz>=0;iz-=j_rndbnd)
       {
          i=iz+ix*nz;
          vel[i]+=(sigma*exp(alpha_sigma*(nztape-iz)))*((2.0*rand()/RAND_MAX)-1.0);
          if (vel[i]>maxV/exp(alpha_maxV*(nztape-iz)))
             vel[i]=maxV/exp(alpha_maxV*(nztape-iz));
          msk[i]=1;
          i=(nz-1-iz)+ix*nz;
          vel[i]+=(sigma*exp(alpha_sigma*(nztape-iz)))*((2.0*rand()/RAND_MAX)-1.0);
          if (vel[i]>maxV/exp(alpha_maxV*(nztape-iz)))
              vel[i]=maxV/exp(alpha_maxV*(nztape-iz));
          msk[i]=1;
       }

    // then adjust the var of the whole boundary regions using nearest neighbor (statistically in all three directions)

    for (ix=0;ix<nx;ix++)
       for (iz=0;iz<nztape;iz++)
       {
          i=nztape-1-iz+ix*nz;
          if (msk[i]==0)
          {
             iz1=((int)(iz/j_rndbnd))*j_rndbnd;
             ix1=((int)(ix/j_rndbnd))*j_rndbnd;
             if ((iz1+j_rndbnd)<nztape)
             {
                wht=1.*(iz-iz1)/j_rndbnd;
                if (((1.0*rand())/RAND_MAX)<=wht)
                   iz1=iz1+j_rndbnd;
             }
             if ((ix1+j_rndbnd)<nx)
             {
                wht=1.*(ix-ix1)/j_rndbnd;
                if (((1.0*rand())/RAND_MAX)<=wht)
                   ix1=ix1+j_rndbnd;
             }
             i1=nztape-1-iz1+ix1*nz;
             vel[i]=vel[i1];
          }

          i=(nz-nztape+iz)+ix*nz;
          if (msk[i]==0)
          {
             iz1=((int)(iz/j_rndbnd))*j_rndbnd;
             ix1=((int)(ix/j_rndbnd))*j_rndbnd;
             if ((iz1+j_rndbnd)<nztape)
             {
                wht=1.*(iz-iz1)/j_rndbnd;
                if (((1.0*rand())/RAND_MAX)<=wht)
                   iz1=iz1+j_rndbnd;
             }
             if ((ix1+j_rndbnd)<nx)
             {
                wht=1.*(ix-ix1)/j_rndbnd;
                if (((1.0*rand())/RAND_MAX)<=wht)
                   ix1=ix1+j_rndbnd;
             }
             i1=(nz-nztape+iz1)+ix1*nz;
             vel[i]=vel[i1];
          }
       }



    for (i=0;i<nx*nz;i++)
    {
       if (vel[i]<minV)
          vel[i]=minV;
    }

   free(msk);
   return 0; 
}


/// n1 n2 n3 refers to model dimension
int XC2dWI_spline_fadj(float *mod, float *dat, int n1, int n2, int n3, int idim, float dmod, int ndat, float ddat)
{
   int i1,i2,i3;
   int imod,idat;
   float x,alpha, beta;
   float coef1,coef2,coef3,coef4;
   int idat1,idat2,idat3,idat4;
   int idx,idx1,idx2,idx3,idx4;
   int imod1,imod2,imod3,imod4;
   int n1dat,n2dat,n3dat;
   int flag1,flag2,flag3,flag4;
   int fat=2;
   n1dat=n1;
   n2dat=n2;
   n3dat=n3;
   if (idim==0)
      n1dat=ndat;
   else if(idim==1)
      n2dat=ndat;
   else if (idim==2)
      n3dat=ndat;
   
   for (i3=0;i3<n3dat;i3++)
      for ( i2=0;i2<n2dat;i2++)
         for (i1=0;i1<n1dat;i1++)
         {
            flag1=0; 
            flag2=0;
            flag3=0;
            flag4=0;
            if (idim==0)
               idat=i1;
            else if(idim==1)
               idat=i2;
            else if (idim==2)
               idat=i3;
 
            x=idat*ddat;
            imod=x/dmod;
            alpha=(x-imod*dmod)/dmod;
            beta=1.-alpha;
            coef1=pow( (2.-(1.+alpha)),3)/6.;
            coef2=(4.-6.*(pow(alpha,2))+3.*(pow(alpha,3)))/6.;
            coef3=(4.-6.*(pow(beta,2))+3.*(pow(beta,3)))/6.;
            coef4=pow((2.-(2.-alpha)),3)/6.;
    
            if (idim==0)
            {
               imod1=imod-1;
               if ( (imod1>-1) && (imod1<n1) )
                  flag1=1;

               imod2=imod;
               if ( (imod2>-1) && (imod2<n1) )
                  flag2=1;

               imod3=imod+1;
               if ( (imod3>-1) && (imod3<n1) )
                  flag3=1;

               imod4=imod+2;
               if ( (imod4>-1) && (imod4<n1) )
                  flag4=1;

               idx1=imod1+i2*n1+i3*n1*n2;
               idx2=imod2+i2*n1+i3*n1*n2;
               idx3=imod3+i2*n1+i3*n1*n2;
               idx4=imod4+i2*n1+i3*n1*n2;
            }
            else if(idim==1)
            {
               imod1=imod-1;
               if ( (imod1>-1) && (imod1<n2) )
                  flag1=1;

               imod2=imod;
               if ( (imod2>-1) && (imod2<n2) )
                  flag2=1;

               imod3=imod+1;
               if ( (imod3>-1) && (imod3<n2) )
                  flag3=1;

               imod4=imod+2;
               if ( (imod4>-1) && (imod4<n2) )
                  flag4=1;

               idx1=i1+imod1*n1+i3*n1*n2;
               idx2=i1+imod2*n1+i3*n1*n2;
               idx3=i1+imod3*n1+i3*n1*n2;
               idx4=i1+imod4*n1+i3*n1*n2;
            }
            else if (idim==2)
            {
               imod1=imod-1;
               if ( (imod1>-1) && (imod1<n3) )
                  flag1=1;

               imod2=imod;
               if ( (imod2>-1) && (imod2<n3) )
                  flag2=1;

               imod3=imod+1;
               if ( (imod3>-1) && (imod3<n3) )
                  flag3=1;

               imod4=imod+2;
               if ( (imod4>-1) && (imod4<n3) )
                  flag4=1;

               idx1=i1+i2*n1+imod1*n1*n2;
               idx2=i1+i2*n1+imod2*n1*n2;
               idx3=i1+i2*n1+imod3*n1*n2;
               idx4=i1+i2*n1+imod4*n1*n2;
            }
            idx=i1+i2*n1dat+i3*n1dat*n2dat;
            if (flag1==1)
               mod[idx1]+=coef1*dat[idx];
            if (flag2==1)
               mod[idx2]+=coef2*dat[idx];
            if (flag3==1) 
               mod[idx3]+=coef3*dat[idx];
            if (flag4==1) 
               mod[idx4]+=coef4*dat[idx]; 
         }  
   return 0;
}

/// n1 n2 n3 refers to model dimension
int XC2dWI_spline_adj(double *mod, double *dat, int n1, int n2, int n3, int idim, float dmod, int ndat, float ddat)
{
   int i1,i2,i3;
   int imod,idat;
   float x,alpha, beta;
   float coef1,coef2,coef3,coef4;
   int idat1,idat2,idat3,idat4;
   int idx,idx1,idx2,idx3,idx4;
   int imod1,imod2,imod3,imod4;
   int n1dat,n2dat,n3dat;
   int flag1,flag2,flag3,flag4;
   int fat=2;
   n1dat=n1;
   n2dat=n2;
   n3dat=n3;
   if (idim==0)
      n1dat=ndat;
   else if(idim==1)
      n2dat=ndat;
   else if (idim==2)
      n3dat=ndat;
   
   for (i3=0;i3<n3dat;i3++)
      for ( i2=0;i2<n2dat;i2++)
         for (i1=0;i1<n1dat;i1++)
         {
            flag1=0; 
            flag2=0;
            flag3=0;
            flag4=0;
            if (idim==0)
               idat=i1;
            else if(idim==1)
               idat=i2;
            else if (idim==2)
               idat=i3;
 
            x=idat*ddat;
            imod=x/dmod;
            alpha=(x-imod*dmod)/dmod;
            beta=1.-alpha;
            coef1=pow( (2.-(1.+alpha)),3)/6.;
            coef2=(4.-6.*(pow(alpha,2))+3.*(pow(alpha,3)))/6.;
            coef3=(4.-6.*(pow(beta,2))+3.*(pow(beta,3)))/6.;
            coef4=pow((2.-(2.-alpha)),3)/6.;
    
            if (idim==0)
            {
               imod1=imod-1;
               if ( (imod1>-1) && (imod1<n1) )
                  flag1=1;

               imod2=imod;
               if ( (imod2>-1) && (imod2<n1) )
                  flag2=1;

               imod3=imod+1;
               if ( (imod3>-1) && (imod3<n1) )
                  flag3=1;

               imod4=imod+2;
               if ( (imod4>-1) && (imod4<n1) )
                  flag4=1;

               idx1=imod1+i2*n1+i3*n1*n2;
               idx2=imod2+i2*n1+i3*n1*n2;
               idx3=imod3+i2*n1+i3*n1*n2;
               idx4=imod4+i2*n1+i3*n1*n2;
               //if (i2<5)
               //  fprintf(stderr,"i2=%d\n",i2);
            }
            else if(idim==1)
            {
               //fprintf(stderr,"idim==1 in adjoint, i1=%d, imod=%d\n",i1,imod);
               imod1=imod-1;
               if ( (imod1>-1) && (imod1<n2) )
                  flag1=1;

               imod2=imod;
               if ( (imod2>-1) && (imod2<n2) )
                  flag2=1;

               imod3=imod+1;
               if ( (imod3>-1) && (imod3<n2) )
                  flag3=1;

               imod4=imod+2;
               if ( (imod4>-1) && (imod4<n2) )
                  flag4=1;

               idx1=i1+imod1*n1+i3*n1*n2;
               idx2=i1+imod2*n1+i3*n1*n2;
               idx3=i1+imod3*n1+i3*n1*n2;
               idx4=i1+imod4*n1+i3*n1*n2;
            }
            else if (idim==2)
            {
               imod1=imod-1;
               if ( (imod1>-1) && (imod1<n3) )
                  flag1=1;

               imod2=imod;
               if ( (imod2>-1) && (imod2<n3) )
                  flag2=1;

               imod3=imod+1;
               if ( (imod3>-1) && (imod3<n3) )
                  flag3=1;

               imod4=imod+2;
               if ( (imod4>-1) && (imod4<n3) )
                  flag4=1;

               idx1=i1+i2*n1+imod1*n1*n2;
               idx2=i1+i2*n1+imod2*n1*n2;
               idx3=i1+i2*n1+imod3*n1*n2;
               idx4=i1+i2*n1+imod4*n1*n2;
            }
            idx=i1+i2*n1dat+i3*n1dat*n2dat;
            /*if (i2==0)
            {
               fprintf(stderr,"idx=%d\n",idx); 
               fprintf(stderr,"idx1=%d\n",idx1); 
               fprintf(stderr,"idx2=%d\n",idx2); 
               fprintf(stderr,"idx3=%d\n",idx3); 
               fprintf(stderr,"idx4=%d\n",idx4); 
            }*/
            if (flag1==1)
               mod[idx1]+=coef1*dat[idx]+0.;
            if (flag2==1)
               mod[idx2]+=coef2*dat[idx]+0.;
            if (flag3==1) 
               mod[idx3]+=coef3*dat[idx]+0.;
            if (flag4==1) 
               mod[idx4]+=coef4*dat[idx]+0.; 
         }  
   return 0;
}
/// n1 n2 n3 refers to model dimension
int XC2dWI_spline_forward(double *mod, double *dat, int n1, int n2, int n3, int idim, float dmod, int ndat, float ddat)
{
   int i1,i2,i3;
   int imod,idat;
   float x,alpha, beta;
   float coef1,coef2,coef3,coef4;
   int idat1,idat2,idat3,idat4;
   int imod1,imod2,imod3,imod4;
   int idx,idx1,idx2,idx3,idx4;
   int n1dat,n2dat,n3dat;
   int flag1,flag2,flag3,flag4;
   int fat=2;
   n1dat=n1;
   n2dat=n2;
   n3dat=n3;
   if (idim==0)
      n1dat=ndat;
   else if(idim==1)
      n2dat=ndat;
   else if (idim==2)
      n3dat=ndat;
   
   for (i3=0;i3<n3dat;i3++)
      for ( i2=0;i2<n2dat;i2++)
         for (i1=0;i1<n1dat;i1++)
         {
            if (idim==0)
               idat=i1;
            else if(idim==1)
               idat=i2;
            else if (idim==2)
               idat=i3;
 
            x=idat*ddat;
            imod=x/dmod;
            alpha=(x-imod*dmod)/dmod;
            beta=1.-alpha;
            coef1=pow( (2.-(1.+alpha)),3)/6.;
            coef2=(4.-6.*(pow(alpha,2))+3.*(pow(alpha,3)))/6.;
            coef3=(4.-6.*(pow(beta,2))+3.*(pow(beta,3)))/6.;
            coef4=pow((2.-(2.-alpha)),3)/6.;

            if (idim==0)
            {
               imod1=min(max(imod-1,0),n1-1);
               imod2=min(max(imod  ,0),n1-1);
               imod3=min(max(imod+1,0),n1-1);
               imod4=min(max(imod+2,0),n1-1);
              
               idx1=imod1+i2*n1+i3*n1*n2;
               idx2=imod2+i2*n1+i3*n1*n2;
               idx3=imod3+i2*n1+i3*n1*n2;
               idx4=imod4+i2*n1+i3*n1*n2;
            }
            else if(idim==1)
            {
               //fprintf(stderr,"idim==1 in forward, i1=%d, imod=%d\n",i1,imod);
               imod1=min(max(imod-1,0),n2-1);
               imod2=min(max(imod  ,0),n2-1);
               imod3=min(max(imod+1,0),n2-1);
               imod4=min(max(imod+2,0),n2-1);
               
               idx1=i1+imod1*n1+i3*n1*n2;
               idx2=i1+imod2*n1+i3*n1*n2;
               idx3=i1+imod3*n1+i3*n1*n2;
               idx4=i1+imod4*n1+i3*n1*n2;
            }
            else if (idim==2)
            {
               imod1=min(max(imod-1,0),n3-1);
               imod2=min(max(imod  ,0),n3-1);
               imod3=min(max(imod+1,0),n3-1);
               imod4=min(max(imod+2,0),n3-1);
               
               idx1=i1+i2*n1+imod1*n1*n2;
               idx2=i1+i2*n1+imod2*n1*n2;
               idx3=i1+i2*n1+imod3*n1*n2;
               idx4=i1+i2*n1+imod4*n1*n2;
            }
            idx=i1+i2*n1dat+i3*n1dat*n2dat;
            dat[idx]+=coef1*mod[idx1]+coef2*mod[idx2]+coef3*mod[idx3]+coef4*mod[idx4];
         }  
   return 0;
}

//int XC3dWI_minterp2v(float *m, float *v, int ndim, int* nm, float* dm, float *om, int* nv, float *dv, float *ov)
